//Ethan Alexander Shulman 2022 "Memento Mori" January 29, 2022

//scene rendering options
#define VIEW_STEPS 40
#define VIEW_STEP_DITHER .1
#define VIEW_EPSILON 1e-4
#define VIEW_EPSILON_MUL 1.1
#define LIGHTING_STEPS 80
#define LIGHTING_STEP_DITHER .1
#define LIGHTING_EPSILON 4e-3
//#define COLOR_MAX 100.
#define DENOISE_RANGE 2
#define DENOISE_NORMAL_THRESHOLD .99
#define DENOISE_DEPTH_THRESHOLD 1e2
#define DENOISE_HASH_DIVIDE 1
//#define ANIMATION_TIME 2.
#define ANIMATION_SAMPLES 1000
#ifdef ANIMATION_TIME
out float time;
#endif


//constants
#define PI 3.14159265


mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))


/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	return mat3(0,0,-5,
			0,0,0,
			3.,3.4,0);
}


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = mod(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy),3.)
#define CHROMA(mt,rc,gc,bc) if (chromaFrame < 1.) {mt[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mt[0]*=vec3(0,3,0);gc;}\
else {mt[0]*=vec3(0,0,3);bc;}



//functions defining scene
float smin(float a, float b, float fadeDst, float smoothness) {
	float d = clamp(.5+(a-b)/fadeDst,0.,1.);//polynomial smooth min from iq http://iquilezles.org/www/articles/smin/smin.htm
	return mix(a,b,d)-fadeDst*d*(1.-d)*smoothness;
}
float smax(float a, float b, float fadeDst, float smoothness) {
	float d = clamp(.5+(a-b)/fadeDst,0.,1.);
	return mix(b,a,d)+fadeDst*d*(1.-d)*smoothness;
}

float sbox(vec3 p, vec3 sz) {
	p = abs(p)-sz;
	return length(max(p,0.))+min(max(p.x,max(p.y,p.z)),0.);
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//skull
	vec3 op = p;
	p.x = abs(p.x);
	vec2 geo = vec2(length(p)-1.,0);//center sphere
	if (geo.x < .8) {
		geo.x = smin(length(p-vec3(.2,.1,.2))-1.,geo.x,.3,.5);//back cranium
		geo.x = smax(-sbox(p-vec3(.3,0,-.82),vec3(.07,.03,.02))+.14,geo.x,0.1,.5);//eye holes
	
		p.xz *= r2d(.5);
		geo.x = min(sbox(p-vec3(-.15,-.6,-.9),vec3(.32,.2,.04)),geo.x);//jaw
		geo.x = max(-(length(p-vec3(-.52,-.3,-.8))-.2), geo.x);//nose
		geo.x = max(-(length(p)-.85),geo.x);//inner cavity
	
		vec3 rp = p;
		rp.x = mod(abs(rp.x)+.03,.06)-.03;
		rp.y = mod(abs(rp.y+.02),.11)-.055;
		rp.z -= -.93;
		float teeth = max(max(p.x+.05,abs(p.y+.58)-.1), sbox(rp,vec3(0.01,0.02,0))-.02);
	
		geo.x = smax(-teeth+.04,geo.x,.1,.3);//hole in jaw for teeth
		geo.x = min(teeth,geo.x);
		
		if (geo.x < .05) {
			#define cwv(d,t,m) max(0.,cos(dot(op,d))-(t))*m
			float bumps = cwv(vec3(16.2,9.5,7.8),-1.,.002)+cwv(vec3(16.,7.,-30.),-1.,.003);//curves/bumps
			geo.x += bumps+cwv(vec3(100,60,40)*.5,.8,5.)*cwv(vec3(130,56,42),0.5,.002)+cwv(vec3(-2,22,14),.95,20.)*max(0.,bumps-.005);//marks/slashes
		}
	} else {
		geo.x -= .7;
	}
	
	//candles
	float odst = max(abs(p.z+1.)-.2,length(op.xy)-3.);
	if (odst < .5) {
		START_VOLUME(p);
	
		vec3 cp = op;
		cp.z -= -1.;
		cp.xy = cos(mod(abs(atan(cp.x,cp.y)),PI/5.)-PI/5.*.5+vec2(0,PI*.5))*length(cp.xy);//circular repetition
		cp.x -= 1.8;
		float cdst = VOLUME(max(length(cp.yz)-.14,abs(cp.x)-.6),.02,.5,0.);//candle sticks
		if (cdst < geo.x) geo = vec2(cdst,1);
	
		float wdst = max(length(cp.yz)-.02,abs(cp.x+.3)-.4);//wicks
		if (wdst < geo.x) geo = vec2(wdst,2);
	
		cp.x -= -.8;
		float flen = length(cp*vec3(1,2.,1)+cos(op.zyx*vec3(23,78,37))*.02),
			fdst = VOLUME(flen-.2,.02,pow(max(0.,1.-flen*5.),4.),0.);//flames
		if (fdst < geo.x) geo = vec2(fdst,3);
	
	} else {
		odst -= .4;
		if (odst < geo.x) geo = vec2(odst,1);
	}
	
	return geo;
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;
float refractionIndex;*/
mat3 material(vec3 p, vec3 n, float mat) {
	if (mat == 1.) {
		//candle sticks
		return mat3(vec3(.9), .06,0,1, 1,0,0);
	}
	if (mat == 2.) {
		//candle wicks
		return mat3(vec3(.002), .1,0,1, 1,0,0);
	}
	if (mat == 3.) {
		//candle flame
		return mat3(vec3(.95,.44,.23), .4,pow(max(0.,p.y*.55+.5),4.)*50.,1, 1,0,0);
	}

	//skull
	float l = clamp(length(mod(abs(p+cos(p.zxy*4.)*.4),.2)-.1)*7.,0.,1.);
	if (baseRand.z < .1) {//mix between glossy and diffuse surface via dithering
		return mat3(vec3(.8), mix(.1,.002,l),0,1, 1,0,0);//outer sheen
	} else {
		return mat3(mix(vec3(.01,.008,.006),vec3(.16,.12,.08),l), .2,0,1, 1,0,0);//dark diffuse surface
	}
}

//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return vec3(0);
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d);
}

